/* 
   Unix SMB/Netbios implementation.
   Version 1.9.
   SMB client smb functions
   Copyright (C) Andrew Tridgell              1994-1998
   Copyright (C) Luke Kenneth Casson Leighton 1996-1998
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifdef SYSLOG
#undef SYSLOG
#endif

#include "includes.h"
#include "trans2.h"

/* value for unused fid field in trans2 secondary request */
#define FID_UNUSED (0xFFFF)

#define CNV_LANG(s) dos2unix_format(s,False)

extern file_info def_finfo;

extern int DEBUGLEVEL;

/****************************************************************************
delete a directory
****************************************************************************/
BOOL cli_rmdir(struct cli_state *cli, int t_idx, char *dname)
{
	char *p;

	cli_set_smb_cmd(cli, t_idx,  SMBrmdir, 0, 2 + strlen(dname),True);

	p = smb_buf(cli->outbuf);
	*p++ = 4;      
	strcpy(p,dname);

	send_smb(cli->fd,cli->outbuf);
	if (!client_receive_smb(cli->fd,cli->inbuf,cli->timeout)) return False;
    if (cli_error(cli, NULL, NULL)) return False;

	return True;
}


/****************************************************************************
delete a file
****************************************************************************/
BOOL cli_unlink(struct cli_state *cli, int t_idx, char *fname)
{
	char *p;

	cli_set_smb_cmd(cli, t_idx,  SMBunlink, 1, 2 + strlen(fname),True);


	SSVAL(cli->outbuf,smb_vwv0,aSYSTEM | aHIDDEN);
  
	p = smb_buf(cli->outbuf);
	*p++ = 4;      
	strcpy(p,fname);

	send_smb(cli->fd,cli->outbuf);
	if (!client_receive_smb(cli->fd,cli->inbuf,cli->timeout)) return False;
    if (cli_error(cli, NULL, NULL)) return False;

	return True;
}


/****************************************************************************
 send a message
 ****************************************************************************/
int cli_send_message(struct cli_state *cli, int t_idx, 
				char *username, char *desthost,
				char *message, int *total_len)
{
	int idx = 0;

	char *p;
	int grp_id;

	if (total_len == NULL) return False;
	*total_len = 0;

	/* send a SMBsendstrt command */
	cli_set_smb_cmd(cli, t_idx,  SMBsendstrt,0,0,True);

	p = smb_buf(cli->outbuf);
	*p++ = 4;
	strcpy(p,username);
	p = skip_string(p,1);
	*p++ = 4;
	strcpy(p,desthost);
	p = skip_string(p,1);

	set_message(cli->outbuf,0,PTR_DIFF(p,smb_buf(cli->outbuf)),False);

	send_smb(cli->fd,cli->outbuf);

	if (!client_receive_smb(cli->fd,cli->inbuf,cli->timeout)) return False;
	if (cli_error(cli, NULL, NULL)) return False;

	grp_id = SVAL(cli->inbuf,smb_vwv0);

	while ((*total_len) < 1600)
	{
		int maxlen = MIN(1600 - (*total_len),127);
		pstring msg;
		int l=0;
		int c;

		bzero(msg, smb_size);

		for (l = 0; l < maxlen && (c = message[idx]) != EOF; l++, idx++)
		{
			msg[l] = c;   
		}

		cli_set_smb_cmd(cli, t_idx,  SMBsendtxt, 1,l+3,True);

		SSVAL(cli->outbuf,smb_vwv0,grp_id);

		p = smb_buf(cli->outbuf);
		*p = 1;
		SSVAL(p,1,l);
		memcpy(p+3,msg,l);

		send_smb(cli->fd,cli->outbuf);

		if (!client_receive_smb(cli->fd,cli->inbuf,cli->timeout)) return False;
		if (cli_error(cli,  NULL, NULL)) return False;

		(*total_len) += l;
	}

	cli_set_smb_cmd(cli, t_idx,  SMBsendend, 1,0,False);
	SSVAL(cli->outbuf,smb_vwv0,grp_id);

	send_smb(cli->fd,cli->outbuf);

	if (!client_receive_smb(cli->fd,cli->inbuf,cli->timeout)) return False;
	if (cli_error(cli,  NULL, NULL)) return False;

	return True;
}

/****************************************************************************
  do a directory listing, calling fn on each file found
  ****************************************************************************/
void cli_do_dir(struct cli_state *cli, int t_idx, struct client_info *info,
				char *Mask,int attribute, BOOL recurse_dir,
				void (*fn)(struct cli_state*, int, struct client_info*, file_info*))
{
  DEBUG(5,("do_dir(mask:%s, attrib:%x, recurse:%s)\n",
				Mask,attribute,BOOLSTR(recurse_dir)));

  if (cli->protocol >= PROTOCOL_LANMAN2)
    {
      if (cli_long_dir(cli, t_idx,  info, Mask, attribute, recurse_dir, fn) > 0)
	return;
    }

  expand_mask(Mask,False);
  cli_short_dir(cli, t_idx,  info, Mask,attribute, recurse_dir, fn);
  return;
}

/*******************************************************************
  decide if a file should be operated on
  ********************************************************************/
static BOOL do_this_one(file_info *finfo, struct client_info *info)
{
  if (finfo->mode & aDIR) return(True);

  if (info->newer_than && finfo->mtime < info->newer_than)
    return(False);

  if ((info->archive_level==1 || info->archive_level==2) && !(finfo->mode & aARCH))
    return(False);

  return(True);
}


/****************************************************************************
  act on the files in a dir listing
  ****************************************************************************/
static void dir_action(struct cli_state *cli, int t_idx, struct client_info *info,
				int attribute, BOOL recurse_dir,
				void (*fn)(struct cli_state*, int, struct client_info*, file_info*),
				file_info *finfo, BOOL longdir)
{

  if (!((finfo->mode & aDIR) == 0 && *info->file_sel && 
	!mask_match(finfo->name,info->file_sel,False,False)) &&
      !(recurse_dir && (strequal(finfo->name,".") || 
			strequal(finfo->name,".."))))
    {
      if (recurse_dir && (finfo->mode & aDIR))
	{
	  pstring mask2;
	  pstring sav_dir;
	  strcpy(sav_dir,info->cur_dir);
	  strcat(info->cur_dir,finfo->name);
	  strcat(info->cur_dir,"\\");
	  strcpy(mask2,info->cur_dir);

	  if (!fn)
	    DEBUG(0,("\n%s\n",CNV_LANG(info->cur_dir)));

	  strcat(mask2,"*");

	  if (longdir)
	    cli_long_dir(cli, t_idx,  info, mask2, attribute, recurse_dir, fn);
	  else
	    cli_dir(cli, t_idx,  info, mask2, attribute, recurse_dir, fn);

	  strcpy(info->cur_dir,sav_dir);
	}
      else
	{
	  if (fn && do_this_one(finfo, info))
	    fn(cli, t_idx,  info, finfo);
	}
    }
}

/****************************************************************************
interpret a long filename structure - this is mostly guesses at the moment
The length of the structure is returned
The structure of a long filename depends on the info level. 260 is used
by NT and 2 is used by OS/2
****************************************************************************/
static int interpret_long_filename(int level,char *p,file_info *finfo,
				struct client_info *info)
{
  if (finfo)
    memcpy(finfo,&def_finfo,sizeof(*finfo));

  switch (level)
    {
    case 1: /* OS/2 understands this */
      if (finfo)
	{
	  /* these dates are converted to GMT by make_unix_date */
	  finfo->ctime = make_unix_date2(p+4);
	  finfo->atime = make_unix_date2(p+8);
	  finfo->mtime = make_unix_date2(p+12);
	  finfo->size = IVAL(p,16);
	  finfo->mode = CVAL(p,24);
	  strcpy(finfo->name,p+27);
	}
      return(28 + CVAL(p,26));

    case 2: /* this is what OS/2 uses mostly */
      if (finfo)
	{
	  /* these dates are converted to GMT by make_unix_date */
	  finfo->ctime = make_unix_date2(p+4);
	  finfo->atime = make_unix_date2(p+8);
	  finfo->mtime = make_unix_date2(p+12);
	  finfo->size = IVAL(p,16);
	  finfo->mode = CVAL(p,24);
	  strcpy(finfo->name,p+31);
	}
      return(32 + CVAL(p,30));

      /* levels 3 and 4 are untested */
    case 3:
      if (finfo)
	{
	  /* these dates are probably like the other ones */
	  finfo->ctime = make_unix_date2(p+8);
	  finfo->atime = make_unix_date2(p+12);
	  finfo->mtime = make_unix_date2(p+16);
	  finfo->size = IVAL(p,20);
	  finfo->mode = CVAL(p,28);
	  strcpy(finfo->name,p+33);
	}
      return(SVAL(p,4)+4);

    case 4:
      if (finfo)
	{
	  /* these dates are probably like the other ones */
	  finfo->ctime = make_unix_date2(p+8);
	  finfo->atime = make_unix_date2(p+12);
	  finfo->mtime = make_unix_date2(p+16);
	  finfo->size = IVAL(p,20);
	  finfo->mode = CVAL(p,28);
	  strcpy(finfo->name,p+37);
	}
      return(SVAL(p,4)+4);

    case 260: /* NT uses this, but also accepts 2 */
      if (finfo)
	{
	  int ret = SVAL(p,0);
	  int namelen;
	  p += 4; /* next entry offset */
	  p += 4; /* fileindex */

	  /* these dates appear to arrive in a weird way. It seems to
	     be localtime plus the serverzone given in the initial
	     connect. This is GMT when DST is not in effect and one
	     hour from GMT otherwise. Can this really be right??

	     I suppose this could be called kludge-GMT. Is is the GMT
	     you get by using the current DST setting on a different
	     localtime. It will be cheap to calculate, I suppose, as
	     no DST tables will be needed */

	  finfo->ctime = interpret_long_date(p); p += 8;
	  finfo->atime = interpret_long_date(p); p += 8;
	  finfo->mtime = interpret_long_date(p); p += 8; p += 8;
	  finfo->size = IVAL(p,0); p += 8;
	  p += 8; /* alloc size */
	  finfo->mode = CVAL(p,0); p += 4;
	  namelen = IVAL(p,0); p += 4;
	  p += 4; /* EA size */
	  p += 2; /* short name len? */
	  p += 24; /* short name? */	  
	  StrnCpy(finfo->name,p,namelen);
	  return(ret);
	}
      return(SVAL(p,0));
    }

  DEBUG(1,("Unknown long filename format %d\n",level));
  return(SVAL(p,0));
}


/****************************************************************************
  get info on a file
  ****************************************************************************/
BOOL cli_stat(struct cli_state *cli, int t_idx, char *file)
{
	BOOL ret;
	pstring param;
	struct mem_buf rparam;
	struct mem_buf rdata ;
	char *p;
	uint16 setup = TRANSACT2_QPATHINFO;

	mem_init(&rparam, 0);
	mem_init(&rdata , 0);

	bzero(param,6);
	SSVAL(param,0,4); /* level */
	p = param+6;
	strcat(p, file);

	cli_send_trans(cli, t_idx, SMBtrans2,NULL,0,FID_UNUSED,0,
	               NULL,param,&setup,
	               0,6 + strlen(p)+1,1,
	               BUFFER_SIZE,2,0);

	ret = cli_receive_trans(cli, t_idx,  SMBtrans2,
	                        &rdata,&rparam);

	mem_free_data(&rparam);
	mem_free_data(&rdata );

	return ret;
}


/****************************************************************************
  display info about a file
  ****************************************************************************/
static void display_finfo(struct cli_state *cli, int t_idx, struct client_info *info,
				file_info *finfo)
{
  if (do_this_one(finfo, info))
  {
    time_t t = finfo->mtime; /* the time is assumed to be passed as GMT */
    DEBUG(0,("  %-30s%7.7s%10d  %s",
  	   CNV_LANG(finfo->name),
	   attrib_string(finfo->mode),
	   finfo->size,
	   asctime(LocalTime(&t))));
	if (info->dir_total)
	{
    	info->dir_total += finfo->size;
	}
  }
}


/****************************************************************************
  do a directory listing, calling fn on each file found. Use the TRANSACT2
  call for long filenames
  ****************************************************************************/
int cli_long_dir(struct cli_state *cli, int t_idx, struct client_info *info,
				char *Mask,int attribute, BOOL recurse_dir,
				void (*fn)(struct cli_state*, int, struct client_info*, file_info*))
{
  int max_matches = 512;
  int info_level = cli->protocol<PROTOCOL_NT1?1:260; /* NT uses 260, OS/2 uses 2. Both accept 1. */
  char *p;
  pstring mask;
  file_info finfo;
  int i;
  char *dirlist = NULL;
  int dirlist_len = 0;
  int total_received = 0;
  BOOL First = True;
	struct mem_buf rparam;
	struct mem_buf rdata ;

  int ff_resume_key = 0;
  int ff_searchcount=0;
  int ff_eos=0;
  int ff_lastname=0;
  int ff_dir_handle=0;
  int loop_count = 0;

  uint16 setup;
  pstring param;

	mem_init(&rparam, 0);
	mem_init(&rdata , 0);

  strcpy(mask,Mask);

  while (ff_eos == 0)
    {
      loop_count++;
      if (loop_count > 200)
	{
	  DEBUG(0,("Error: Looping in FIND_NEXT??\n"));
	  break;
	}

      if (First)
	{
	  setup = TRANSACT2_FINDFIRST;
	  SSVAL(param,0,attribute); /* attribute */
	  SSVAL(param,2,max_matches); /* max count */
	  SSVAL(param,4,8+4+2);	/* resume required + close on end + continue */
	  SSVAL(param,6,info_level); 
	  SIVAL(param,8,0);
	  strcpy(param+12,mask);
	}
      else
	{
	  setup = TRANSACT2_FINDNEXT;
	  SSVAL(param,0,ff_dir_handle);
	  SSVAL(param,2,max_matches); /* max count */
	  SSVAL(param,4,info_level); 
	  SIVAL(param,6,ff_resume_key); /* ff_resume_key */
	  SSVAL(param,10,8+4+2);	/* resume required + close on end + continue */
	  strcpy(param+12,mask);

	  DEBUG(5,("hand=0x%X resume=%d ff_lastname=%d mask=%s\n",
		   ff_dir_handle,ff_resume_key,ff_lastname,mask));
	}
      /* ??? original code added 1 pad byte after param */

      cli_send_trans(cli, t_idx, SMBtrans2,NULL,0,FID_UNUSED,0,
			 NULL,param,&setup,
			 0,12+strlen(mask)+1,1,
			 BUFFER_SIZE,10,0);

      if (!cli_receive_trans(cli, t_idx,  SMBtrans2,
			          &rdata,&rparam))
	{
	  DEBUG(3,("FIND%s gave %s\n",First?"FIRST":"NEXT", cli_errstr(cli)));
	  break;
	}

      /* parse out some important return info */
      p = rparam.data;
      if (First)
	{
	  ff_dir_handle = SVAL(p,0);
	  ff_searchcount = SVAL(p,2);
	  ff_eos = SVAL(p,4);
	  ff_lastname = SVAL(p,8);
	}
      else
	{
	  ff_searchcount = SVAL(p,0);
	  ff_eos = SVAL(p,2);
	  ff_lastname = SVAL(p,6);
	}

      if (ff_searchcount == 0) 
	break;

      /* point to the data bytes */
      p = rdata.data;

      /* we might need the lastname for continuations */
      if (ff_lastname > 0)
	{
	  switch(info_level)
	    {
	    case 260:
	      ff_resume_key =0;
	      StrnCpy(mask,p+ff_lastname, rdata.data_used-ff_lastname);
	      /* strcpy(mask,p+ff_lastname+94); */
	      break;
	    case 1:
	      strcpy(mask,p + ff_lastname + 1);
	      ff_resume_key = 0;
	      break;
	    }
	}
      else
	strcpy(mask,"");
  
      /* and add them to the dirlist pool */
      dirlist = Realloc(dirlist,dirlist_len +  rdata.data_used);

      if (!dirlist)
	{
	  DEBUG(0,("Failed to expand dirlist\n"));
	  break;
	}

      /* put in a length for the last entry, to ensure we can chain entries 
	 into the next packet */
      {
	char *p2;
	for (p2=p,i=0;i<(ff_searchcount-1);i++)
	  p2 += interpret_long_filename(info_level,p2,NULL, info);
	SSVAL(p2,0, rdata.data_used - PTR_DIFF(p2,p));
      }

      /* grab the data for later use */
      memcpy(dirlist+dirlist_len,p, rdata.data_used);
      dirlist_len +=  rdata.data_used;

      total_received += ff_searchcount;

	mem_free_data(&rparam);
	mem_free_data(&rdata );

      DEBUG(3,("received %d entries (eos=%d resume=%d)\n",
	       ff_searchcount,ff_eos,ff_resume_key));

      First = False;
    }

  if (!fn)
    for (p=dirlist,i=0;i<total_received;i++)
      {
	p += interpret_long_filename(info_level,p,&finfo, info);
	display_finfo(cli, t_idx,  info, &finfo);
      }

  for (p=dirlist,i=0;i<total_received;i++)
    {
      p += interpret_long_filename(info_level,p,&finfo, info);
      dir_action(cli, t_idx,  info, attribute, recurse_dir, fn, &finfo, True);
    }

  /* free up the dirlist buffer */
  if (dirlist) free(dirlist);
  return(total_received);
}


/****************************************************************************
interpret a short filename structure
The length of the structure is returned
****************************************************************************/
static int interpret_short_filename(char *p,file_info *finfo)
{
  finfo->mode = CVAL(p,21);

  /* this date is converted to GMT by make_unix_date */
  finfo->ctime = make_unix_date(p+22);
  finfo->mtime = finfo->atime = finfo->ctime;
  finfo->size = IVAL(p,26);
  strcpy(finfo->name,p+30);
  
  return(DIR_STRUCT_SIZE);
}

/****************************************************************************
  do a directory listing, calling fn on each file found
  ****************************************************************************/
int cli_short_dir(struct cli_state *cli, int t_idx, struct client_info *info,
				char *Mask,int attribute, BOOL recurse_dir,
				void (*fn)(struct cli_state*, int, struct client_info*, file_info*))
{
  char *p;
  int received = 0;
  BOOL first = True;
  char status[21];
  int num_asked = (cli->max_xmit - 100)/DIR_STRUCT_SIZE;
  int num_received = 0;
  int i;
  char *dirlist = NULL;
  pstring mask;
  file_info finfo;

  finfo = def_finfo;

  bzero(status,21);

  strcpy(mask,Mask);
  
  while (1)
    {
		int num_bytes = 0;
		int num_words = 0;
		uint8 cmd = 0;

      bzero(cli->outbuf,smb_size);
      if (first)	
		{
			num_words = 2;
			num_bytes = 5 + strlen(mask);
		}
      else
		{
			num_words = 2;
			num_bytes = 5 + 21;
		}

#if FFIRST
      if (cli->protocol >= PROTOCOL_LANMAN1)
	cmd = SMBffirst;
      else
#endif
	cmd = SMBsearch;

	cli_set_smb_cmd(cli, t_idx,  cmd, num_words, num_bytes, True);

      SSVAL(cli->outbuf,smb_vwv0,num_asked);
      SSVAL(cli->outbuf,smb_vwv1,attribute);
  
      p = smb_buf(cli->outbuf);
      *p++ = 4;
      
      if (first)
	strcpy(p,mask);
      else
	strcpy(p,"");
      p += strlen(p) + 1;
      
      *p++ = 5;
      if (first)
	SSVAL(p,0,0);
      else
	{
	  SSVAL(p,0,21);
	  p += 2;
	  memcpy(p,status,21);
	}

	send_smb(cli->fd,cli->outbuf);
	if (client_receive_smb(cli->fd, cli->inbuf, cli->timeout))
	{ 
      received = SVAL(cli->inbuf,smb_vwv0);
    }
    else
    {
      received = 0;
    }

      DEBUG(5,("dir received %d\n",received));

      DEBUG(6,("errstr=%s\n",cli_errstr(cli)));

      if (received <= 0) break;

      first = False;

      dirlist = Realloc(dirlist,(num_received + received)*DIR_STRUCT_SIZE);

      if (!dirlist) 
	return 0;

      p = smb_buf(cli->inbuf) + 3;

      memcpy(dirlist+num_received*DIR_STRUCT_SIZE,
	     p,received*DIR_STRUCT_SIZE);

      memcpy(status,p + ((received-1)*DIR_STRUCT_SIZE),21);

      num_received += received;

      if (CVAL(cli->inbuf,smb_rcls) != 0) break;
    }

#if FFIRST
  if (!first && cli->protocol >= PROTOCOL_LANMAN1)
    {
      bzero(cli->outbuf,smb_size);
      CVAL(cli->outbuf,smb_com) = SMBfclose;

      SSVAL(cli->outbuf,smb_tid,cli->con[t_idx].cnum );
      cli_setup_packet(cli);

      p = smb_buf(cli->outbuf);
      *p++ = 4;
      
      strcpy(p,"");
      p += strlen(p) + 1;
      
      *p++ = 5;
      SSVAL(p,0,21);
      p += 2;
      memcpy(p,status,21);

	send_smb(cli->fd,cli->outbuf);
	if (!client_receive_smb(cli->fd, cli->inbuf, cli->timeout) ||
	     cli_error(cli,  NULL, NULL))
	{ 
	DEBUG(0,("Error closing search: %s\n", cli_errstr(cli)));      
    }
#endif

  if (!fn)
    for (p=dirlist,i=0;i<num_received;i++)
      {
	p += interpret_short_filename(p,&finfo);
	display_finfo(cli, t_idx,  info, &finfo);
      }

  for (p=dirlist,i=0;i<num_received;i++)
    {
      p += interpret_short_filename(p,&finfo);
      dir_action(cli, t_idx,  info, attribute, recurse_dir, fn, &finfo, False);
    }

  if (dirlist) free(dirlist);
  return(num_received);
}

/****************************************************************************
  do a directory listing, calling fn on each file found
  ****************************************************************************/
void cli_dir(struct cli_state *cli, int t_idx, struct client_info *info,
				char *Mask,int attribute, BOOL recurse_dir,
				void (*fn)(struct cli_state*, int, struct client_info*, file_info*))
{
  DEBUG(5,("cli_dir(%s,%x,%s)\n",Mask,attribute,BOOLSTR(recurse_dir)));
  if (cli->protocol >= PROTOCOL_LANMAN2)
    {
      if (cli_long_dir(cli, t_idx,  info, Mask, attribute, recurse_dir, fn) > 0)
	return;
    }

  expand_mask(Mask,False);
  cli_short_dir(cli, t_idx,  info, Mask, attribute, recurse_dir, fn);
  return;
}


/****************************************************************************
  put a single file
  ****************************************************************************/
int cli_put(struct cli_state *cli, int t_idx, struct client_info *info,
				char *rname,char *lname, file_info *finfo,
				int (*read_fn)(struct client_info *, char*, int, int, FILE *))
{
  uint16 fnum;
  FILE *f;
  int nread=0;
  time_t close_time = finfo->mtime;
  char *buf=NULL;
  static int maxwrite=0;

  struct timeval tp_start;
  GetTimeOfDay(&tp_start);

  f = fopen(lname,"r");

  if (!f)
    {
      DEBUG(0,("cli_put: Error opening local file %s\n", lname));
      return False;
    }

  if (finfo->mtime == 0 || finfo->mtime == -1)
    finfo->mtime = finfo->atime = finfo->ctime = time(NULL);

  if (!cli_create(cli, t_idx,  rname, finfo->mode, finfo->mtime, &fnum)) return False;

  if (fnum == 0xffff) return False;

  if (finfo->size < 0)
    finfo->size = file_size(lname);
  
  DEBUG(1,("putting file %s of size %d bytes as %s\n",
		lname,finfo->size,CNV_LANG(rname)));
  
  if (!maxwrite)
    maxwrite = cli->writebraw_supported?MAX(cli->max_xmit,BUFFER_SIZE):(cli->max_xmit-200);

  while (nread < finfo->size)
    {
      int n = maxwrite;
      int ret;

      n = MIN(n,finfo->size - nread);

      /* HACK! this is for *possible* use by writeraw */
      buf = (char *)Realloc(buf,n+4);
  
      fseek(f,nread,SEEK_SET);
      /* buf+4 because that's where the data is */
      if ((n = read_fn(info, buf+4,1,n, f)) < 1)
	{
	  DEBUG(0,("Error reading local file\n"));
	  break;
	}	  

      /* buf+4 is the data.  buf[0..3] *might* get used by writeraw */
      ret = cli_write(cli, t_idx,  fnum, nread, buf+4, n);

      DEBUG(5,("cli_write: offset %ld requested %d.  received %d\n",
			nread, n, ret));

      if (n != ret) {
	if (!maxwrite) {
	  DEBUG(0,("Error writing file\n"));
	  break;
	} else {
	  maxwrite /= 2;
	  continue;
	}
      }

      nread += n;
    }

    if (!cli_close(cli, t_idx,  fnum, close_time))
    {
      DEBUG(0,("%s closing remote file %s\n",smb_errstr(cli->inbuf),CNV_LANG(rname)));
      fclose(f);
      if (buf) free(buf);
      return False;
    }

  
  fclose(f);
  if (buf) free(buf);

  {
    struct timeval tp_end;
    int this_time;

    GetTimeOfDay(&tp_end);
    this_time = 
      (tp_end.tv_sec - tp_start.tv_sec)*1000 +
	(tp_end.tv_usec - tp_start.tv_usec)/1000;
    info->put_total_time_ms += this_time;
    info->put_total_size += finfo->size;

    DEBUG(1,("(%g kb/s) (average %g kb/s)\n",
	     finfo->size          / (1.024*this_time + 1.0e-4),
	     info->put_total_size / (1.024*info->put_total_time_ms)));
  }

  return True;
} 

 
/****************************************************************************
  get a file from rname to lname
  ****************************************************************************/
int cli_get(struct cli_state *cli, int t_idx, struct client_info *info,
				char *rname,char *lname,file_info *finfo1,
				int handle,
				int (*init_fn)(struct client_info *, int, char*, file_info *),
				int (*write_fn)(struct client_info *, int, char*, int),
				int (*end_fn)(struct client_info*, int, char*, int, file_info*))
{  
  uint32 nread=0;
  char *p;
  file_info finfo;
  BOOL close_done = False;
  BOOL ignore_close_error = False;
  char *dataptr=NULL;
  int datalen=0;

  uint16 fnum;
  uint16 fmode;
  time_t mtime;
  uint32 fsize;

	struct timeval tar_tp_start;
	GetTimeOfDay(&tar_tp_start);

  if (finfo1) 
    finfo = *finfo1;
  else
    finfo = def_finfo;

  if (info->lowercase)
    strlower(lname);

  fnum = cli_open(cli, t_idx,  rname, O_RDONLY, DENY_NONE, &fmode, &mtime, &fsize);
			
  if (fnum == 0xffff) return False;

  strcpy(finfo.name,rname);

  if (!finfo1)
    {
      finfo.mode = fmode;
      /* these times arrive as LOCAL time, using the DST offset 
	 corresponding to that time, we convert them to GMT */
      finfo.atime = finfo.ctime = finfo.mtime = mtime;
      finfo.size = fsize;
    }

  /* following a successful cli_create(), call the initialisation function */
  if (init_fn)
  {
	if (!init_fn(info, handle, rname, &finfo)) return 0;
  }

  /* we might have got some data from a chained readX */
  if (SVAL(cli->inbuf,smb_vwv0) == SMBreadX)
    {
      p = (smb_base(cli->inbuf)+SVAL(cli->inbuf,smb_vwv1)) - smb_wct;
      datalen = SVAL(p,smb_vwv5);
      dataptr = smb_base(cli->inbuf) + SVAL(p,smb_vwv6);
    }
  else
    {
      dataptr = NULL;
      datalen = 0;
    }


  DEBUG(2,("getting file %s of size %d bytes as %s ",
	   CNV_LANG(finfo.name),
	   finfo.size,
	   lname));

  while (nread < finfo.size && !close_done)
    {
      int method = -1;
      static BOOL can_chain_close = True;

      p=NULL;
      
      DEBUG(3,("nread=%d max_xmit=%d fsize=%d\n",nread,cli->max_xmit,finfo.size));

      /* 3 possible read types. readbraw if a large block is required.
	 readX + close if not much left and read if neither is supported */

      /* we might have already read some data from a chained readX */
      if (dataptr && datalen>0)
	method=3;

      /* if we can finish now then readX+close */
      if (method<0 && can_chain_close && (cli->protocol >= PROTOCOL_LANMAN1) && 
	  ((finfo.size - nread) < 
	   (cli->max_xmit - (2*smb_size + 13*SIZEOFWORD + 300))))
	method = 0;

      /* if we support readraw then use that */
      if (method<0 && cli->readbraw_supported)
	method = 1;

      /* if we can then use readX */
      if (method<0 && (cli->protocol >= PROTOCOL_LANMAN1))
	method = 2;

      switch (method)
	{
	  /* use readX */
	case 0:
	case 2:
	  if (method == 0)
	    close_done = True;
	    
	  /* use readX + close */
	cli_set_smb_cmd(cli, t_idx,  SMBreadX, 10,0,True);
	  
	  if (close_done)
	    {
	      CVAL(cli->outbuf,smb_vwv0) = SMBclose;
	      SSVAL(cli->outbuf,smb_vwv1,smb_offset(smb_buf(cli->outbuf),cli->outbuf));
	    }
	  else
	    CVAL(cli->outbuf,smb_vwv0) = 0xFF;	      
	  
	  SSVAL(cli->outbuf,smb_vwv2,fnum);
	  SIVAL(cli->outbuf,smb_vwv3,nread);
	  SSVAL(cli->outbuf,smb_vwv5,MIN(cli->max_xmit-200,finfo.size - nread));
	  SSVAL(cli->outbuf,smb_vwv6,0);
	  SIVAL(cli->outbuf,smb_vwv7,0);
	  SSVAL(cli->outbuf,smb_vwv9,MIN(BUFFER_SIZE,finfo.size-nread));
	  
	  if (close_done)
	    {
	      p = smb_buf(cli->outbuf);
	      bzero(p,9);
	      
	      CVAL(p,0) = 3;
	      SSVAL(p,1,fnum);
	      SIVALS(p,3,-1);
	      
	      /* now set the total packet length */
	      smb_setlen(cli->outbuf,smb_len(cli->outbuf)+9);
	    }
	  
      send_smb(cli->fd,cli->outbuf);
      if (!client_receive_smb(cli->fd, cli->inbuf, cli->timeout)) break;
      if (cli_error(cli,  NULL, NULL)) break;

	  if (close_done &&
	      SVAL(cli->inbuf,smb_vwv0) != SMBclose)
	    {
	      /* NOTE: WfWg sometimes just ignores the chained
		 command! This seems to break the spec? */
	      DEBUG(3,("Rejected chained close?\n"));
	      close_done = False;
	      can_chain_close = False;
	      ignore_close_error = True;
	    }
	  
	  datalen = SVAL(cli->inbuf,smb_vwv5);
	  dataptr = smb_base(cli->inbuf) + SVAL(cli->inbuf,smb_vwv6);
	  break;

	  /* use readbraw */
	case 1:
	  {
	    static int readbraw_size = BUFFER_SIZE;
	  
	cli_set_smb_cmd(cli, t_idx,  SMBreadbraw, 8,0,True);

	    SSVAL(cli->outbuf,smb_vwv0,fnum);
	    SIVAL(cli->outbuf,smb_vwv1,nread);
	    SSVAL(cli->outbuf,smb_vwv3,MIN(finfo.size-nread,readbraw_size));
	    SSVAL(cli->outbuf,smb_vwv4,0);
	    SIVALS(cli->outbuf,smb_vwv5,-1);

      send_smb(cli->fd,cli->outbuf);

	    /* Now read the raw data into the buffer and write it */	  
	    if(read_smb_length(cli->fd,cli->inbuf,0) == -1) {
	      DEBUG(0,("Failed to read length in readbraw\n"));	    
	      exit(1);
	    }
	    
	    /* Even though this is not an smb message, smb_len
	       returns the generic length of an smb message */
	    datalen = smb_len(cli->inbuf);

	    if (datalen == 0)
	      {
		/* we got a readbraw error */
		DEBUG(4,("readbraw error - reducing size\n"));
		readbraw_size = (readbraw_size * 9) / 10;
		
		if (readbraw_size < cli->max_xmit)
		  {
		    DEBUG(0,("disabling readbraw\n"));
		    cli->readbraw_supported = False;
		  }
		
		dataptr=NULL;
		continue;
	      }

	    if(read_data(cli->fd,cli->inbuf,datalen) != datalen) {
	      DEBUG(0,("Failed to read data in readbraw\n"));
	      exit(1);
	    }
	    dataptr = cli->inbuf;
	  }
	  break;

	case 3:
	  /* we've already read some data with a chained readX */
	  break;

	default:
	  /* use plain read */

	cli_set_smb_cmd(cli, t_idx,  SMBread, 5,0,True);

	  SSVAL(cli->outbuf,smb_vwv0,fnum);
	  SSVAL(cli->outbuf,smb_vwv1,MIN(cli->max_xmit-200,finfo.size - nread));
	  SIVAL(cli->outbuf,smb_vwv2,nread);
	  SSVAL(cli->outbuf,smb_vwv4,finfo.size - nread);

  send_smb(cli->fd,cli->outbuf);
  if (!client_receive_smb(cli->fd, cli->inbuf, cli->timeout)) break;
  if (cli_error(cli,  NULL, NULL)) break;

	  datalen = SVAL(cli->inbuf,smb_vwv0);
	  dataptr = smb_buf(cli->inbuf) + 3;
	  break;
	}
 
      if (write_fn(info, handle,dataptr,datalen) != datalen)
	{
	  DEBUG(0,("Error writing local file\n"));
	  break;
	}
      
      nread += datalen;
      if (datalen == 0) 
	{
	  DEBUG(0,("Error reading file %s. Got %d bytes\n",CNV_LANG(rname),nread));
	  break;
	}

      dataptr=NULL;
      datalen=0;
    }

    if (end_fn)
	{
       end_fn(info, handle, cli->inbuf, nread, &finfo);
    }

  if (!close_done)
    {

      cli_close(cli, t_idx,  fnum, 0);
      
      if (!ignore_close_error && cli_error(cli,  NULL, NULL))
	{
	  DEBUG(0,("Error %s closing remote file\n",cli_errstr(cli)));
	  return 0;
	}
    }

	{
		struct timeval tar_tp_end;
		int this_time;

		GetTimeOfDay(&tar_tp_end);
		this_time = 
		(tar_tp_end.tv_sec - tar_tp_start.tv_sec)*1000 +
		(tar_tp_end.tv_usec - tar_tp_start.tv_usec)/1000;
		info->get_total_time_ms += this_time;
		info->get_total_size    += nread;

		/* Thanks to Carel-Jan Engel (ease@mail.wirehub.nl) for this one */
		DEBUG(1,("(%g kb/s) (average %g kb/s)\n",
			nread                / MAX(0.001, (1.024*this_time)),
			info->get_total_size / MAX(0.001,(1.024*info->get_total_time_ms))));
	}

  return nread;
}

/****************************************************************************
check for existance of a dir
****************************************************************************/
BOOL cli_chkpath(struct cli_state *cli, int t_idx, char *path)
{
	fstring path2;
	char *p;

	strcpy(path2,path);
	trim_string(path2,NULL,"\\");
	if (!*path2) *path2 = '\\';

	cli_set_smb_cmd(cli, t_idx,  SMBchkpth, 0,4 + strlen(path2),True);

	p = smb_buf(cli->outbuf);
	*p++ = 4;
	strcpy(p,path2);

#if 0
	{
	/* this little bit of code can be used to extract NT error codes.
	   Just feed a bunch of "cd foo" commands to smbclient then watch
	   in netmon (tridge)
	 */
		static int code=0;
		SIVAL(outbuf, smb_rcls, code | 0xC0000000);
		SSVAL(outbuf, smb_flg2, SVAL(outbuf, smb_flg2) | (1<<14));
		code++;
	}
#endif

	send_smb(cli->fd,cli->outbuf);
	if (!client_receive_smb(cli->fd,cli->inbuf,cli->timeout))
	{
		return False;
	}

    if (cli_error(cli, NULL, NULL)) return False;

	return True;
}


/****************************************************************************
check the space on a device
****************************************************************************/
BOOL cli_dskattr(struct cli_state *cli, int t_idx,
				uint16 *num_blocks, uint32 *block_size, uint16 *free_blocks)
{
	cli_set_smb_cmd(cli, t_idx,  SMBdskattr, 0,0,True);

	send_smb(cli->fd,cli->outbuf);
	if (!client_receive_smb(cli->fd,cli->inbuf,cli->timeout)) return False;
    if (cli_error(cli, NULL, NULL)) return False;

	*num_blocks  = SVAL(cli->inbuf,smb_vwv0);
	*block_size  = SVAL(cli->inbuf,smb_vwv1) * SVAL(cli->inbuf,smb_vwv2);
	*free_blocks = SVAL(cli->inbuf,smb_vwv3);

	return True;
}

/****************************************************************************
make a directory of name "name"
****************************************************************************/
BOOL cli_mkdir(struct cli_state *cli, int t_idx, char *name)
{
	char *p;

	cli_set_smb_cmd(cli, t_idx,  SMBmkdir, 0,2 + strlen(name),True);

	p = smb_buf(cli->outbuf);
	*p++ = 4;      
	strcpy(p,name);
  
	send_smb(cli->fd,cli->outbuf);
	if (!client_receive_smb(cli->fd,cli->inbuf,cli->timeout)) return False;
    if (cli_error(cli, NULL, NULL)) return False;

	return(True);
}


/****************************************************************************
rename some files
****************************************************************************/
BOOL cli_move(struct cli_state *cli, int t_idx, char *src, char *dest)
{
	char *p;

	cli_set_smb_cmd(cli, t_idx,  SMBmv, 1, 4 + strlen(src) + strlen(dest),True);

	SSVAL(cli->outbuf,smb_vwv0,aHIDDEN | aDIR | aSYSTEM);

	p = smb_buf(cli->outbuf);
	*p++ = 4;      
	strcpy(p,src);
	p = skip_string(p,1);
	*p++ = 4;      
	strcpy(p,dest);

	send_smb(cli->fd,cli->outbuf);
	if (!client_receive_smb(cli->fd,cli->inbuf,cli->timeout)) return False;
    if (cli_error(cli, NULL, NULL)) return False;

	return(True);
}

/****************************************************************************
Get DOS file attributes
***************************************************************************/
BOOL cli_getatr(struct cli_state *cli, int t_idx, char *fname,
				uint16 *fattr, time_t *ftime, uint32 *fsize)
{
	char *p;

	cli_set_smb_cmd(cli, t_idx,  SMBgetatr, 0,2 + strlen(fname),True);

	p = smb_buf(cli->outbuf);
	*p++ = 4;
	strcpy(p,fname);
	p += (strlen(fname)+1);

	*p++ = 4;
	*p++ = 0; /* zero this, due to byte-alignment issues... */

	send_smb(cli->fd,cli->outbuf);
	if (!client_receive_smb(cli->fd,cli->inbuf,cli->timeout)) return False;
	if (cli_error(cli,  NULL, NULL)) return False;

	if (fattr) (*fattr) = SVAL(cli->inbuf,smb_vwv0);
	if (ftime) make_unix_date3(cli->inbuf + smb_vwv1);
	if (fsize) (*fsize) = IVAL(cli->inbuf,smb_vwv3);

	if (fattr && ftime && fsize)
	{
		DEBUG(5,("SMBgetatr attr:0x%x time:%d size:%d\n",
				*fattr, *ftime, *fsize));
	}

	return(True);
}


/****************************************************************************
Set DOS file attributes
***************************************************************************/
BOOL cli_setatr(struct cli_state *cli, int t_idx, char *fname,
				uint16 fattr, time_t write_time)
{
	char *p;

	cli_set_smb_cmd(cli, t_idx,  SMBsetatr, 8,4 + strlen(fname),True);

	SSVAL(cli->outbuf,smb_vwv0, fattr);
	put_dos_date3(cli->outbuf,smb_vwv1, write_time); /* zero indicates no change */
	/* all the other 6 words are reserved, so says cifs6.txt... */

	p = smb_buf(cli->outbuf);
	*p++ = 4;      
	strcpy(p,fname);
	p += (strlen(fname)+1);

	*p++ = 4;
	*p++ = 0; /* zero this, due to byte-alignment issues... */

	send_smb(cli->fd,cli->outbuf);
	if (!client_receive_smb(cli->fd,cli->inbuf,cli->timeout)) return False;
	if (cli_error(cli,  NULL, NULL)) return False;

	return(True);
}


/****************************************************************************
send a qpathinfo call

c_time - create time
a_time - accessed time
m_time - modified time

****************************************************************************/
BOOL cli_qpathinfo(struct cli_state *cli, uint16 t_idx, char *fname, 
		   time_t *c_time, time_t *a_time, time_t *m_time, uint32 *size)
{
	uint16 setup = TRANSACT2_QPATHINFO;
	pstring param;
	struct mem_buf rparam;
	struct mem_buf rdata ;

	int param_len = strlen(fname) + 7;

	mem_init(&rparam, 0);
	mem_init(&rdata , 0);

	memset(param, 0, param_len);
	SSVAL(param, 0, SMB_INFO_STANDARD);
	pstrcpy(&param[6], fname);

	if (cli_api(cli, t_idx, param_len, 0, 1,
		    1024,BUFFER_SIZE,
		    param,NULL, &setup,
		    &rparam,&rdata)) {
		return False;
	}

	if (!rdata.data || rdata.data_used < 22) {
		return False;
	}

	if (c_time) (*c_time) = make_unix_date2(rdata.data+0);
	if (a_time) (*a_time) = make_unix_date2(rdata.data+4);
	if (m_time) (*m_time) = make_unix_date2(rdata.data+8);
	if (size  ) (*size  ) = IVAL(rdata.data, 12);

	mem_free_data(&rparam);
	mem_free_data(&rdata );

	return True;
}

/****************************************************************************
send a qpathinfo call with the SMB_QUERY_FILE_ALL_INFO info level

c_time - create time
a_time - accessed time
m_time - modified time
w_time - write time

****************************************************************************/
BOOL cli_qpathinfo2(struct cli_state *cli, uint16 t_idx, char *fname, 
		    time_t *c_time, time_t *a_time, time_t *m_time, 
		    time_t *w_time, uint32 *size)
{
	uint16 setup = TRANSACT2_QPATHINFO;
	pstring param;
	struct mem_buf rparam;
	struct mem_buf rdata ;

	int param_len = strlen(fname) + 7;

	mem_init(&rparam, 0);
	mem_init(&rdata , 0);

	memset(param, 0, param_len);
	SSVAL(param, 0, SMB_QUERY_FILE_ALL_INFO);
	pstrcpy(&param[6], fname);

	if (cli_api(cli, t_idx, param_len, 0, 1,
		    1024,BUFFER_SIZE,
		    param,NULL, &setup,
		    &rparam,&rdata)) {
		return False;
	}

	if (!rdata.data || rdata.data_used < 22) {
		return False;
	}

	if (c_time) (*c_time) = interpret_long_date(rdata.data+ 0) - cli->serverzone;
	if (a_time) (*a_time) = interpret_long_date(rdata.data+ 8) - cli->serverzone;
	if (m_time) (*m_time) = interpret_long_date(rdata.data+16) - cli->serverzone;
	if (w_time) (*w_time) = interpret_long_date(rdata.data+24) - cli->serverzone;
	if (size  ) (*size  ) = IVAL(rdata.data, 40);

	mem_free_data(&rparam);
	mem_free_data(&rdata );

	return True;
}


/****************************************************************************
send a qfileinfo call

c_time - create time
a_time - accessed time
m_time - modified time

****************************************************************************/
BOOL cli_qfileinfo(struct cli_state *cli, uint16 t_idx, uint16 fnum, 
		   time_t *c_time, time_t *a_time, time_t *m_time, uint32 *size)
{
	uint16 setup = TRANSACT2_QFILEINFO;
	pstring param;
	struct mem_buf rparam;
	struct mem_buf rdata ;

	int param_len = 4;

	mem_init(&rparam, 0);
	mem_init(&rdata , 0);

	memset(param, 0, param_len);
	SSVAL(param, 0, fnum);
	SSVAL(param, 2, SMB_INFO_STANDARD);

	if (cli_api(cli, t_idx,  param_len,0,1,
		    1024,BUFFER_SIZE,
		    param,NULL, &setup,
		    &rparam,&rdata)) {
		return False;
	}

	if (!rdata.data || rdata.data_used < 22) {
		return False;
	}

	if (c_time) (*c_time) = make_unix_date2(rdata.data+0);
	if (a_time) (*a_time) = make_unix_date2(rdata.data+4);
	if (m_time) (*m_time) = make_unix_date2(rdata.data+8);
	if (size  ) (*size  ) = IVAL(rdata.data, 12);

	mem_free_data(&rparam);
	mem_free_data(&rdata );

	return True;
}

/****************************************************************************
Create a file on a share
***************************************************************************/
BOOL cli_create(struct cli_state *cli, int t_idx,
				char *name, uint16 file_mode, uint16 make_time, uint16 *fnum)
{
	char *p;

	cli_set_smb_cmd(cli, t_idx,  SMBcreate, 3,2 + strlen(name),True);

	SSVAL(cli->outbuf,smb_vwv0,file_mode);
	put_dos_date3(cli->outbuf,smb_vwv1,make_time);

	p = smb_buf(cli->outbuf);
	*p++ = 4;      
	strcpy(p, name);

	send_smb(cli->fd,cli->outbuf);
	if (!client_receive_smb(cli->fd,cli->inbuf,cli->timeout)) return False;
	if (cli_error(cli,  NULL, NULL)) return False;

	*fnum = SVAL(cli->inbuf,smb_vwv0);
	return True;
}

/****************************************************************************
open a file
****************************************************************************/
uint16 cli_open(struct cli_state *cli, int t_idx, char *fname, int flags, int share_mode,
			uint16 *fmode, time_t *mtime, uint32 *fsize)
{
	char *p;
	unsigned openfn=0;
	unsigned accessmode=0;
	uint16 fnum;

	DEBUG(5,("cli_open: cli fd:%d t_idx:%d fname:%s\n", cli->fd, t_idx, fname));

	if (flags & O_CREAT)
	{
		openfn |= (1<<4);
	}

	if (!(flags & O_EXCL))
	{
		if (flags & O_TRUNC)
		{
			openfn |= (1<<1);
		}
		else
		{
			openfn |= (1<<0);
		}
	}

	accessmode = (share_mode<<4);

	if (IS_BITS_SET_ALL(flags, O_RDWR))
	{
		accessmode |= 4;
	}
	else if (IS_BITS_SET_ALL(flags, O_WRONLY))
	{
		accessmode |= 2;
	} 
	else
	{
		accessmode |= 1;
	} 

	cli_set_smb_cmd(cli, t_idx,  SMBopenX, 15,1 + strlen(fname),True);

	SSVAL(cli->outbuf,smb_vwv0,0xFF);
	SSVAL(cli->outbuf,smb_vwv2,1);  /* return additional info */
	SSVAL(cli->outbuf,smb_vwv3,accessmode);
	SSVAL(cli->outbuf,smb_vwv4,aSYSTEM | aHIDDEN);
	SSVAL(cli->outbuf,smb_vwv5,0);
	SSVAL(cli->outbuf,smb_vwv8,openfn);
	SSVAL(cli->outbuf,smb_vwv11,0xffff);
	SSVAL(cli->outbuf,smb_vwv12,0xffff);
  
	p = smb_buf(cli->outbuf);
	strcpy(p,fname);
	p = skip_string(p,1);

	send_smb(cli->fd,cli->outbuf);
	if (!client_receive_smb(cli->fd,cli->inbuf,cli->timeout)) return 0xffff;
    if (cli_error(cli,  NULL, NULL)) return 0xffff;

	fnum                = SVAL(cli->inbuf,smb_vwv2);

    if (fmode) (*fmode) = SVAL(cli->inbuf,smb_vwv3);
    if (mtime) (*mtime) = make_unix_date3(cli->inbuf+smb_vwv4);
    if (fsize) (*fsize) = IVAL(cli->inbuf,smb_vwv6);

	DEBUG(5,("cli_open: opening fnum:%04x\n", fnum));

	return fnum;
}


/****************************************************************************
  close a file
****************************************************************************/
BOOL cli_close(struct cli_state *cli, int t_idx, uint16 fnum, time_t close_time)
{
	cli_set_smb_cmd(cli, t_idx,  SMBclose, 3,0,True); 

	SSVAL(cli->outbuf,smb_vwv0,fnum);
	put_dos_date3(cli->outbuf,smb_vwv1, close_time);
	SSVAL(cli->outbuf,smb_vwv2,0);

	send_smb(cli->fd,cli->outbuf);
	if (!client_receive_smb(cli->fd,cli->inbuf,cli->timeout)) {
		return False;
	}

    if (cli_error(cli, NULL, NULL)) return False;

	return True;
}


/****************************************************************************
  lock a file
****************************************************************************/
BOOL cli_lock(struct cli_state *cli, int t_idx, uint16 fnum, uint32 offset, uint32 len, int timeout)
{
	char *p;

	cli_set_smb_cmd(cli, t_idx,  SMBlockingX, 8,10,True);

	CVAL(cli->outbuf,smb_vwv0) = 0xFF;
	SSVAL(cli->outbuf,smb_vwv2,fnum);
	CVAL(cli->outbuf,smb_vwv3) = 0;
	SIVALS(cli->outbuf, smb_vwv4, timeout);
	SSVAL(cli->outbuf,smb_vwv6,0);
	SSVAL(cli->outbuf,smb_vwv7,1);

	p = smb_buf(cli->outbuf);
	SSVAL(p, 0, cli->pid);
	SIVAL(p, 2, offset);
	SIVAL(p, 6, len);

	send_smb(cli->fd,cli->outbuf);
	if (!client_receive_smb(cli->fd,cli->inbuf,cli->timeout)) {
		return False;
	}

    if (cli_error(cli, NULL, NULL)) return False;

	return True;
}

/****************************************************************************
  unlock a file
****************************************************************************/
BOOL cli_unlock(struct cli_state *cli, int t_idx, uint16 fnum, uint32 offset, uint32 len, int timeout)
{
	char *p;

	cli_set_smb_cmd(cli, t_idx,  SMBlockingX, 8,10,True);

	CVAL(cli->outbuf,smb_vwv0) = 0xFF;
	SSVAL(cli->outbuf,smb_vwv2,fnum);
	CVAL(cli->outbuf,smb_vwv3) = 0;
	SIVALS(cli->outbuf, smb_vwv4, timeout);
	SSVAL(cli->outbuf,smb_vwv6,1);
	SSVAL(cli->outbuf,smb_vwv7,0);

	p = smb_buf(cli->outbuf);
	SSVAL(p, 0, cli->pid);
	SIVAL(p, 2, offset);
	SIVAL(p, 6, len);

	send_smb(cli->fd,cli->outbuf);
	if (!client_receive_smb(cli->fd,cli->inbuf,cli->timeout)) {
		return False;
	}

    if (cli_error(cli, NULL, NULL)) return False;

	return True;
}


/****************************************************************************
  read from a file
****************************************************************************/
int cli_readx(struct cli_state *cli, int t_idx, uint16 fnum, char *buf, uint32 offset, uint16 size)
{
	char *p;

	DEBUG(5,("cli_readx: fnum:%4x offset:%d size:%d\n", fnum, offset, size));

	cli_set_smb_cmd(cli, t_idx,  SMBreadX, 12,0,True);

	CVAL(cli->outbuf,smb_vwv0) = 0xFF;
	SSVAL(cli->outbuf,smb_vwv2,fnum); /* file number */
	SIVAL(cli->outbuf,smb_vwv3,offset); /* file offset */
	SSVAL(cli->outbuf,smb_vwv5,size); /* max size count */
	SSVAL(cli->outbuf,smb_vwv6,size); /* min size count */
	SIVAL(cli->outbuf,smb_vwv7,-1); /* open timeout */
	SSVAL(cli->outbuf,smb_vwv9,0); /* bytes left */
	SIVAL(cli->outbuf,smb_vwv10,0); /* file offset, high (64 bit size?) */
	SSVAL(cli->outbuf,smb_vwv11,0); /* bytes count */

	show_msg(cli->outbuf);
	send_smb(cli->fd,cli->outbuf);
	if (!client_receive_smb(cli->fd,cli->inbuf,cli->timeout)) return -1;
	show_msg(cli->inbuf);

    if (cli_error(cli, NULL, NULL)) return -1;

	size = SVAL(cli->inbuf, smb_vwv5);
	p = smb_base(cli->inbuf) + SVAL(cli->inbuf,smb_vwv6);

	memcpy(buf, p, size);

	DEBUG(5,("cli_readx: returned size:%d\n", size));

	return size;
}


/*******************************************************************
  write to a file using writebraw
  ********************************************************************/
int cli_writeraw(struct cli_state *cli, int t_idx, uint16 fnum,int pos,char *buf,int n)
{
	DEBUG(5,("cli_write_raw: [%04x] %d %d\n", fnum, pos, n));

	cli_set_smb_cmd(cli, t_idx,  SMBwritebraw, cli->protocol>PROTOCOL_COREPLUS?12:10,0,True);

	SSVAL(cli->outbuf,smb_vwv0,fnum);
	SSVAL(cli->outbuf,smb_vwv1,n);
	SIVAL(cli->outbuf,smb_vwv3,pos);
	SSVAL(cli->outbuf,smb_vwv7,1);

	send_smb(cli->fd,cli->outbuf);
	if (!client_receive_smb(cli->fd,cli->inbuf,cli->timeout)) return -1;
    if (cli_error(cli, NULL, NULL)) return -1;

	/* direct write */
	_smb_setlen(buf-4,n);		/* HACK! XXXX */
	if (write_socket(cli->fd,buf-4,n+4) != n+4) return(0);

	if (!client_receive_smb(cli->fd,cli->inbuf,cli->timeout)) return -1;

	return(SVAL(cli->inbuf,smb_vwv0));
}
      


/*******************************************************************
  write to a file
  ********************************************************************/
int cli_write(struct cli_state *cli, int t_idx, uint16 fnum,int pos,char *buf,int n)
{
	int len;

	if (cli->writebraw_supported && n > (cli->max_xmit-200)) 
	{
		return(cli_writeraw(cli, t_idx,  fnum, pos, buf, n));
	}

	cli_set_smb_cmd(cli, t_idx,  SMBwrite, 5,n + 3,True);

	SSVAL(cli->outbuf,smb_vwv0,fnum);
	SSVAL(cli->outbuf,smb_vwv1,n);
	SIVAL(cli->outbuf,smb_vwv2,pos);
	SSVAL(cli->outbuf,smb_vwv4,0);
	CVAL(smb_buf(cli->outbuf),0) = 1;
	SSVAL(smb_buf(cli->outbuf),1,n);

	memcpy(smb_buf(cli->outbuf)+3,buf,n);

	send_smb(cli->fd,cli->outbuf);
	if (!client_receive_smb(cli->fd,cli->inbuf,cli->timeout)) return -1;
    if (cli_error(cli, NULL, NULL)) return -1;

	len = SVAL(cli->inbuf,smb_vwv0);

	DEBUG(5,("cli_write: [%04x] offset:%ld req:%d ret:%d\n",
				fnum, pos, n, len));

	return len;
}
      


/****************************************************************************
  write to a file
****************************************************************************/
int cli_write_x(struct cli_state *cli, int t_idx, uint16 fnum, char *buf, uint32 offset, uint16 size)
{
	char *p;

	bzero(cli->outbuf,smb_size);
	bzero(cli->inbuf,smb_size);

	cli_set_smb_cmd(cli, t_idx,  SMBwriteX, 12,size,True);

	CVAL(cli->outbuf,smb_vwv0) = 0xFF;
	SSVAL(cli->outbuf,smb_vwv2,fnum);
	SIVAL(cli->outbuf,smb_vwv3,offset);

	SSVAL(cli->outbuf,smb_vwv10,size);
	SSVAL(cli->outbuf,smb_vwv11,smb_buf(cli->outbuf) - smb_base(cli->outbuf));

	p = smb_base(cli->outbuf) + SVAL(cli->outbuf,smb_vwv11);
	memcpy(p, buf, size);

	send_smb(cli->fd,cli->outbuf);
	if (!client_receive_smb(cli->fd,cli->inbuf,cli->timeout)) return -1;
    if (cli_error(cli, NULL, NULL)) return -1;

	return SVAL(cli->inbuf, smb_vwv2);
}

