/* 
   Unix SMB/Netbios implementation.
   Version 1.9.
   SMB client generic functions
   Copyright (C) Andrew Tridgell              1994-1998
   Copyright (C) Luke Kenneth Casson Leighton 1996-1998
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifdef SYSLOG
#undef SYSLOG
#endif

#include "includes.h"
#include "trans2.h"

/* value for unused fid field in trans2 secondary request */
#define FID_UNUSED (0xFFFF)

#define CNV_LANG(s) dos2unix_format(s,False)

extern file_info def_finfo;

extern int DEBUGLEVEL;

/****************************************************************************
  send a SMB trans or trans2 request
  ****************************************************************************/
BOOL cli_send_trans(struct cli_state *cli, int t_idx,
			   int trans, char *name, int pipe_name_len,
               int fid, int flags,
			   char *data,char *param,uint16 *setup, int ldata,int lparam,
			   int lsetup,int mdata,int mparam,int msetup)
{
	int i;
	int this_ldata,this_lparam;
	int tot_data=0,tot_param=0;
	char *outdata,*outparam;
	char *p;

	this_lparam = MIN(lparam,cli->max_xmit - (500+lsetup*2)); /* hack */
	this_ldata = MIN(ldata,cli->max_xmit - (500+lsetup*2+this_lparam));

	cli_set_smb_cmd(cli, t_idx,  trans, 14+lsetup, 0, True);

	outparam = smb_buf(cli->outbuf)+(trans==SMBtrans ? (pipe_name_len+1) : 3);
	outdata = outparam+this_lparam;

	/* primary request */
	SSVAL(cli->outbuf,smb_tpscnt,lparam);	/* tpscnt */
	SSVAL(cli->outbuf,smb_tdscnt,ldata);	/* tdscnt */
	SSVAL(cli->outbuf,smb_mprcnt,mparam);	/* mprcnt */
	SSVAL(cli->outbuf,smb_mdrcnt,mdata);	/* mdrcnt */
	SCVAL(cli->outbuf,smb_msrcnt,msetup);	/* msrcnt */
	SSVAL(cli->outbuf,smb_flags,flags);	/* flags */
	SIVAL(cli->outbuf,smb_timeout,0);		/* timeout */
	SSVAL(cli->outbuf,smb_pscnt,this_lparam);	/* pscnt */
	SSVAL(cli->outbuf,smb_psoff,smb_offset(outparam,cli->outbuf)); /* psoff */
	SSVAL(cli->outbuf,smb_dscnt,this_ldata);	/* dscnt */
	SSVAL(cli->outbuf,smb_dsoff,smb_offset(outdata,cli->outbuf)); /* dsoff */
	SCVAL(cli->outbuf,smb_suwcnt,lsetup);	/* suwcnt */
	for (i=0;i<lsetup;i++)		/* setup[] */
		SSVAL(cli->outbuf,smb_setup+i*2,setup[i]);
	p = smb_buf(cli->outbuf);
	if (trans==SMBtrans) {
		memcpy(p,name, pipe_name_len + 1);			/* name[] */
	} else {
		*p++ = 0;  /* put in a null smb_name */
		*p++ = 'D'; *p++ = ' ';	/* observed in OS/2 */
	}
	if (this_lparam)			/* param[] */
		memcpy(outparam,param,this_lparam);
	if (this_ldata)			/* data[] */
		memcpy(outdata,data,this_ldata);
	set_message(cli->outbuf,14+lsetup,		/* wcnt, bcc */
		    PTR_DIFF(outdata+this_ldata,smb_buf(cli->outbuf)),False);

	show_msg(cli->outbuf);
	send_smb(cli->fd,cli->outbuf);

	if (this_ldata < ldata || this_lparam < lparam) {
		/* receive interim response */
		if (!client_receive_smb(cli->fd,cli->inbuf,cli->timeout) || 
		    cli_error(cli, NULL, NULL)) {
			return(False);
		}      

		tot_data = this_ldata;
		tot_param = this_lparam;
		
		while (tot_data < ldata || tot_param < lparam)  {
			this_lparam = MIN(lparam-tot_param,cli->max_xmit - 500); /* hack */
			this_ldata = MIN(ldata-tot_data,cli->max_xmit - (500+this_lparam));

			set_message(cli->outbuf,trans==SMBtrans?8:9,0,True);
			CVAL(cli->outbuf,smb_com) = trans==SMBtrans ? SMBtranss : SMBtranss2;
			
			outparam = smb_buf(cli->outbuf);
			outdata = outparam+this_lparam;
			
			/* secondary request */
			SSVAL(cli->outbuf,smb_tpscnt,lparam);	/* tpscnt */
			SSVAL(cli->outbuf,smb_tdscnt,ldata);	/* tdscnt */
			SSVAL(cli->outbuf,smb_spscnt,this_lparam);	/* pscnt */
			SSVAL(cli->outbuf,smb_spsoff,smb_offset(outparam,cli->outbuf)); /* psoff */
			SSVAL(cli->outbuf,smb_spsdisp,tot_param);	/* psdisp */
			SSVAL(cli->outbuf,smb_sdscnt,this_ldata);	/* dscnt */
			SSVAL(cli->outbuf,smb_sdsoff,smb_offset(outdata,cli->outbuf)); /* dsoff */
			SSVAL(cli->outbuf,smb_sdsdisp,tot_data);	/* dsdisp */
			if (trans==SMBtrans2)
				SSVALS(cli->outbuf,smb_sfid,fid);		/* fid */
			if (this_lparam)			/* param[] */
				memcpy(outparam,param,this_lparam);
			if (this_ldata)			/* data[] */
				memcpy(outdata,data,this_ldata);
			set_message(cli->outbuf,trans==SMBtrans?8:9, /* wcnt, bcc */
				    PTR_DIFF(outdata+this_ldata,smb_buf(cli->outbuf)),False);
			
			show_msg(cli->outbuf);
			send_smb(cli->fd,cli->outbuf);
			
			tot_data += this_ldata;
			tot_param += this_lparam;
		}
	}

	return(True);
}


/****************************************************************************
  receive a SMB trans or trans2 response allocating the necessary memory
  ****************************************************************************/
BOOL cli_receive_trans(struct cli_state *cli, int t_idx, int trans,
				 struct mem_buf *data, struct mem_buf *param)
{
	int data_len  = 0;
	int param_len = 0;

	int total_data  = 0;
	int total_param = 0;

	int this_data;
	int this_param;
	
	data ->data_used = 0;
	param->data_used = 0;
	
	if (!client_receive_smb(cli->fd,cli->inbuf,cli->timeout)) return False;

	show_msg(cli->inbuf);
	
	/* sanity check */
	if (CVAL(cli->inbuf,smb_com) != trans) {
		DEBUG(0,("Expected %s response, got command 0x%02x\n",
			 trans==SMBtrans?"SMBtrans":"SMBtrans2", 
			 CVAL(cli->inbuf,smb_com)));
		return(False);
	}
	if (cli_error(cli, NULL, NULL)) return(False);

	/* parse out the lengths */
	total_data  = SVAL(cli->inbuf,smb_tdrcnt);
	total_param = SVAL(cli->inbuf,smb_tprcnt);

	/* allocate it */
	mem_realloc_data(data , total_data);
	mem_realloc_data(param, total_param);

	while (1)
	{
		this_data = SVAL(cli->inbuf,smb_drcnt);
		this_param = SVAL(cli->inbuf,smb_prcnt);

		if (this_data + data_len > total_data ||
		    this_param + param_len > total_param)
		{
			DEBUG(1,("Data overflow in cli_receive_trans\n"));
			return False;
		}

		if (this_data)
		{
			memcpy(data->data + SVAL(cli->inbuf,smb_drdisp),
			       smb_base(cli->inbuf) + SVAL(cli->inbuf,smb_droff),
			       this_data);
		}
		if (this_param)
		{
			memcpy(param->data + SVAL(cli->inbuf,smb_prdisp),
			       smb_base(cli->inbuf) + SVAL(cli->inbuf,smb_proff),
			       this_param);
		}

		data_len  += this_data;
		param_len += this_param;

		/* parse out the total lengths again - they can shrink! */
		total_data  = SVAL(cli->inbuf,smb_tdrcnt);
		total_param = SVAL(cli->inbuf,smb_tprcnt);
		
		if (total_data <= data_len && total_param <= param_len)
			break;
		
		if (!client_receive_smb(cli->fd,cli->inbuf,cli->timeout))
			return False;

		show_msg(cli->inbuf);
		
		/* sanity check */
		if (CVAL(cli->inbuf,smb_com) != trans) {
			DEBUG(0,("Expected %s response, got command 0x%02x\n",
				 trans==SMBtrans?"SMBtrans":"SMBtrans2", 
				 CVAL(cli->inbuf,smb_com)));
			return(False);
		}
		if (cli_error(cli, NULL, NULL)) return(False);
	}
	
	DEBUG(5,("cli_receive_trans: data len: %d param_len: %d\n",
	          data_len, param_len));

	data ->data_used = data_len ;
	param->data_used = param_len;

	return(True);
}

/****************************************************************************
call a remote api on an arbitrary pipe.  takes param, data and setup buffers.

parameters:

  char *pipe_name, int pipe_name_len,       pipe name, length
  int prcnt,int drcnt, int srcnt,           param, data, setup sizes
  int mprcnt,int mdrcnt,                    max param, data return sizes
  int *rparam.data_used,int *rdata.data_used,                  actual param, data return sizes
  char *param, char *data, uint16 *setup,   params, data, setup buffers
  char **rparam,char **rdata                return params, data buffers

****************************************************************************/
BOOL cli_api_pipe(struct cli_state *cli, int t_idx,
	char *pipe_name, int pipe_name_len,
	uint32 prcnt,uint32 drcnt, uint32 srcnt,
	uint32 mprcnt,uint32 mdrcnt,
	char *param, char *data, uint16 *setup,
	struct mem_buf *rparam, struct mem_buf *rdata)
{
  if (pipe_name_len == 0) pipe_name_len = strlen(pipe_name);

  cli_send_trans(cli, t_idx,  SMBtrans, pipe_name, pipe_name_len, 0,0,
		     data, param, setup,
		     drcnt, prcnt, srcnt,
		     mdrcnt, mprcnt, 0);

  return (cli_receive_trans(cli, t_idx, SMBtrans, rdata,rparam));
}

/****************************************************************************
call a remote api on the LANMAN pipe.  only takes param and data buffers
****************************************************************************/
BOOL cli_api(struct cli_state *cli, int t_idx,
		    int prcnt,int drcnt, int srcnt,
			int mprcnt,int mdrcnt,
			char *param,char *data, uint16 *setup,
			struct mem_buf *rparam, struct mem_buf *rdata)
{
	return cli_api_pipe(cli, t_idx,  "\\PIPE\\LANMAN", 0,
				prcnt, drcnt, srcnt,
				mprcnt, mdrcnt,
				param, data, setup,
				rparam, rdata);
}


/****************************************************************************
perform a NetWkstaUserLogon
****************************************************************************/
BOOL cli_NetWkstaUserLogon(struct cli_state *cli, int t_idx,char *user, char *workstation)
{
	struct mem_buf rparam;
	struct mem_buf rdata ;
	char *p;
	pstring param;

	mem_init(&rparam, 0);
	mem_init(&rdata , 0);

	memset(param, 0, sizeof(param));
	
	/* send a SMBtrans command with api NetWkstaUserLogon */
	p = param;
	SSVAL(p,0,132); /* api number */
	p += 2;
	strcpy(p,"OOWb54WrLh");
	p = skip_string(p,1);
	strcpy(p,"WB21BWDWWDDDDDDDzzzD");
	p = skip_string(p,1);
	SSVAL(p,0,1);
	p += 2;
	strcpy(p,user);
	strupper(p);
	p += 21; p++; p += 15; p++; 
	strcpy(p, workstation); 
	strupper(p);
	p += 16;
	SSVAL(p, 0, BUFFER_SIZE);
	p += 2;
	SSVAL(p, 0, BUFFER_SIZE);
	p += 2;
	
	cli->error = -1;
	
	if (cli_api(cli, t_idx,  PTR_DIFF(p,param),0,0,
		    1024,BUFFER_SIZE,
		    param,NULL,NULL,
		    &rparam,&rdata)) {
		cli->error = SVAL(rparam.data,0);
		p = rdata.data;
		
		if (cli->error == 0) {
			DEBUG(4,("NetWkstaUserLogon success\n"));
			cli->privileges = SVAL(p, 24);
			fstrcpy(cli->eff_name,p+2);
		} else {
			DEBUG(1,("NetwkstaUserLogon gave error %d\n", cli->error));
		}
	}
	
	mem_free_data(&rparam);
	mem_free_data(&rdata );
	return cli->error == 0;
}


/****************************************************************************
try and browse available connections on a host
****************************************************************************/
BOOL cli_NetShareEnum(struct cli_state *cli, int t_idx,
				FILE* hnd, enum display_type display, enum action_type action, 
				BOOL sort, BOOL *long_share_name,
				void (*fn)(FILE *, enum display_type, enum action_type, char *, uint32, char *))
{
#ifdef NOSTRCASECMP
/* If strcasecmp is already defined, remove it. */
#ifdef strcasecmp
#undef strcasecmp
#endif /* strcasecmp */
#define strcasecmp StrCaseCmp
#endif /* NOSTRCASECMP */

	extern int strcasecmp();

	struct mem_buf rparam;
	struct mem_buf rdata ;
	char *p;
	pstring param;
	int count = -1;

	mem_init(&rparam, 0);
	mem_init(&rdata , 0);

	/* now send a SMBtrans command with api RNetShareEnum */
	p = param;
	SSVAL(p,0,0); /* api number */
	p += 2;
	strcpy(p,"WrLeh");
	p = skip_string(p,1);
	strcpy(p,"B13BWz");
	p = skip_string(p,1);
	SSVAL(p,0,1);
	SSVAL(p,2,BUFFER_SIZE);
	p += 4;

	if (cli_api(cli, t_idx,  
		    PTR_DIFF(p,param), /* param count */
		    0, /*data count */
			0, /* setup count */
		    10, /* mprcount */
		    BUFFER_SIZE, /* mdrcount */
		    param, NULL, NULL,
	        &rparam,    &rdata))
	{
		int res = SVAL(rparam.data,0);
		int converter=SVAL(rparam.data,2);
		int i;

		DEBUG(5,("cli_NetShareEnum: res %d conv: %d\n", res, converter));

		if (res == 0)
		{
			count=SVAL(rparam.data,4);
			p = rdata.data;

			if (sort) qsort(p,count,20,QSORT_CAST strcasecmp);

			for (i = 0; i < count; i++)
			{
				char *sname = p;
				uint32 type = SVAL(p,14);
				int comment_offset = IVAL(p,16) & 0xFFFF;
				char *comment = NULL;

				if (comment_offset)
				{
					comment = rdata.data+comment_offset-converter;
				}

				fn(hnd, display, action, sname, type, comment);

				if (long_share_name && strlen(sname) > 8) *long_share_name=True;

				p += 20;
			}
		}
	}

	mem_free_data(&rparam);
	mem_free_data(&rdata );

	return(count>0);
}


/****************************************************************************
call a NetServerEnum for the specified workgroup and servertype mask.
This function then calls the specified callback function for each name returned.

The callback function takes 3 arguments: the machine name, the server type and
the comment.
****************************************************************************/
BOOL cli_NetServerEnum(struct cli_state *cli, int t_idx,
				FILE* hnd, enum display_type display, enum action_type action, 
				char *workgroup, uint32 stype,
				void (*fn)(FILE *, enum display_type, enum action_type, char *, uint32, char *))
{
	struct mem_buf rparam;
	struct mem_buf rdata ;
	char *p;
	pstring param;
	int uLevel = 1;
	int count = -1;
  
	mem_init(&rparam, 0);
	mem_init(&rdata , 0);

	/* send a SMBtrans command with api NetServerEnum */
	p = param;
	SSVAL(p,0,0x68); /* api number */
	p += 2;
	strcpy(p,"zWrLeh");
	p = skip_string(p,1);
  
	strcpy(p,"B21");
  
	p = skip_string(p,1);
	SSVAL(p,0,uLevel);
	SSVAL(p,2,BUFFER_SIZE);
	p += 4;
	
	if (cli_api(cli, t_idx,  
		    PTR_DIFF(p,param), /* param count */
		    0, /*data count */
			0, /* setup count */
		    8, /* mprcount */
		    BUFFER_SIZE, /* mdrcount */
		    param, NULL, NULL,
		    &rparam,&rdata)) {
		int res = SVAL(rparam.data,0);
		int converter=SVAL(rparam.data,2);
		int i;
			
		if (res == 0) {
			count=SVAL(rparam.data,4);
			p = rdata.data;
					
			for (i = 0;i < count;i++, p += 26) {
				char *sname = p;
				int comment_offset = (IVAL(p,22) & 0xFFFF)-converter;
				char *cmnt = comment_offset?(rdata.data+comment_offset):"";
				if (comment_offset < 0 || comment_offset > rdata.data_used) continue;

				stype = IVAL(p,18) & ~SV_TYPE_LOCAL_LIST_ONLY;

				if (fn) fn(hnd, display, action, sname, stype, cmnt);
			}
		}
	}
  
	mem_free_data(&rparam);
	mem_free_data(&rdata );
	
	return(count > 0);
}




/****************************************************************************
call a NetLookupNames 
****************************************************************************/
BOOL cli_NetLookupNames(struct cli_state *cli, int t_idx,
				FILE* hnd, enum display_type display, enum action_type action, 
				char *name,
				void (*fn)(FILE *, enum display_type, enum action_type, char *))
{
	struct mem_buf rparam;
	struct mem_buf rdata ;
	char *p;
	pstring param;
	int uLevel = 0;
	int count = -1;
  
	mem_init(&rparam, 0);
	mem_init(&rdata , 0);

	/* send a SMBtrans command with api NetServerEnum */
	p = param;
	SSVAL(p,0,0x34); /* api number */
	p += 2;

	strcpy(p,"zWrLeh");
	p = skip_string(p,1);
  
	strcpy(p,"B21");
	p = skip_string(p,1);

	pstrcpy(p, name);
	p = skip_string(p,1);
	
	SSVAL(p,0,uLevel);
	SSVAL(p,2,BUFFER_SIZE);
	p += 4;
	
	if (cli_api(cli, t_idx,  
		    PTR_DIFF(p,param), /* param count */
		    0, /*data count */
			0, /* setup count */
		    8, /* mprcount */
		    BUFFER_SIZE, /* mdrcount */
		    param, NULL, NULL,
		    &rparam,&rdata))
	{
		int res = SVAL(rparam.data,0);
		int converter=SVAL(rparam.data,2);
		int i;
			
		if (res == 0)
		{
			count=SVAL(rparam.data,4);
			p = rdata.data;
					
			for (i = 0;i < count;i++, p += 21)
			{
				char *sname = p;
				if (fn) fn(hnd, display, action, sname);
			}
		}
	}
  
	mem_free_data(&rparam);
	mem_free_data(&rdata );
	
	return(count > 0);
}



